# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2022 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Elmer equation object Heat"
__author__ = "Markus Hovorka, Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from femtools import femutils
from . import nonlinear
from ... import equationbase

CONVECTION_TYPE = ["None", "Computed", "Constant"]
PHASE_CHANGE_MODEL = ["None", "Spatial 1", "Spatial 2", "Temporal"]


def create(doc, name="Heat"):
    return femutils.createObject(doc, name, Proxy, ViewProxy)


class Proxy(nonlinear.Proxy, equationbase.HeatProxy):

    Type = "Fem::EquationElmerHeat"

    def __init__(self, obj):
        super().__init__(obj)

        # according to the Elmer models manual Bubbles is by default True
        # and Stabilize is False (Stabilize is added in linear.py)
        obj.addProperty("App::PropertyBool", "Bubbles", "Heat", "", locked=True)
        obj.addProperty(
            "App::PropertyEnumeration",
            "Convection",
            "Equation",
            "Type of convection to be used",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "PhaseChangeModel",
            "Equation",
            "Model for phase change",
            locked=True,
        )

        obj.Bubbles = True
        obj.Stabilize = False
        obj.Convection = CONVECTION_TYPE
        obj.Convection = "None"
        obj.PhaseChangeModel = PHASE_CHANGE_MODEL
        obj.PhaseChangeModel = "None"
        obj.Priority = 20


class ViewProxy(nonlinear.ViewProxy, equationbase.HeatViewProxy):
    pass


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Elmer equation object StaticCurrent"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from femtools import femutils
from ... import equationbase
from . import nonlinear


def create(doc, name="StaticCurrent"):
    return femutils.createObject(doc, name, Proxy, ViewProxy)


class Proxy(nonlinear.Proxy, equationbase.StaticCurrentProxy):

    Type = "Fem::EquationElmerStaticCurrent"

    def __init__(self, obj):
        super().__init__(obj)

        obj.addProperty(
            "App::PropertyBool", "CalculateVolumeCurrent", "StaticCurrent", "", locked=True
        )
        obj.CalculateVolumeCurrent = True
        obj.addProperty(
            "App::PropertyBool", "CalculateJouleHeating", "StaticCurrent", "", locked=True
        )
        obj.addProperty(
            "App::PropertyBool",
            "ConstantWeights",
            "StaticCurrent",
            "Used to turn constant weighting on for the results",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyBool",
            "CalculateNodalHeating",
            "StaticCurrent",
            "Calculate nodal heating that may be used to couple the heat equation optimally when using conforming finite element meshes",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyBool",
            "HeatSource",
            "StaticCurrent",
            "Use Joule heating as a heat source in combination with heat equation",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyBool",
            "PowerControl",
            "StaticCurrent",
            "Apply power control with the desired heating power",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyBool",
            "CurrentControl",
            "StaticCurrent",
            "Apply current control with the desired current",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyElectricCurrent",
            "Current",
            "StaticCurrent",
            "Current control value",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyPower", "Power", "StaticCurrent", "Power control value", locked=True
        )


class ViewProxy(nonlinear.ViewProxy, equationbase.StaticCurrentViewProxy):
    pass


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

import glob
import os

import Arch_rc
import PartGui

# populate the list of submodules
modules = glob.glob(os.path.join(os.path.dirname(__file__), "*.py"))
__all__ = [
    os.path.basename(f)[:-3] for f in modules if os.path.isfile(f) and not f.endswith("__init__.py")
]

from . import *



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import pathlib
from typing import Optional, Tuple, Type, Iterable
from PySide.QtWidgets import QFileDialog, QMessageBox
from ...camassets import cam_assets
from ..manager import AssetManager
from ..serializer import AssetSerializer, Asset
from .util import (
    make_import_filters,
    make_export_filters,
    get_serializer_from_extension,
)
import Path
import Path.Preferences as Preferences


class AssetOpenDialog(QFileDialog):
    def __init__(
        self,
        asset_manager: AssetManager,
        asset_class: Type[Asset],
        serializers: Iterable[Type[AssetSerializer]],
        parent=None,
    ):
        super().__init__(parent)

        # Set default directory based on asset type
        default_dir = self._get_default_directory(asset_class)
        self.setDirectory(default_dir.as_posix())

        self.asset_class = asset_class
        self.asset_manager = asset_manager
        self.serializers = list(serializers)
        self.setFileMode(QFileDialog.ExistingFile)
        filters = make_import_filters(self.serializers)
        self.setNameFilters(filters)
        if filters:
            self.selectNameFilter(filters[0])  # Default to "All supported files"

    def deserialize_file(self, file_path: pathlib.Path, quiet=False) -> Optional[Asset]:
        """Deserialize the selected file using the appropriate serializer."""
        # Find the correct serializer for the file.
        file_extension = file_path.suffix.lower()
        serializer_class = get_serializer_from_extension(
            self.serializers, file_extension, for_import=True
        )
        if not serializer_class:
            message = f"No supported serializer found for file extension '{file_extension}'"
            if quiet:
                Path.Log.error(message)
            else:
                QMessageBox.critical(
                    self,
                    "Error",
                    message,
                )
            return None

        # Check whether all dependencies for importing the file exist.
        try:
            raw_data = file_path.read_bytes()
            dependencies = serializer_class.extract_dependencies(raw_data)
            external_toolbits = []  # Track toolbits found externally

            for dependency_uri in dependencies:
                # First check if dependency exists in asset manager stores
                if self.asset_manager.exists(dependency_uri, store=["local", "builtin"]):
                    continue

                # If not in stores, check if it exists relative to the library file
                dependency_found = False
                if dependency_uri.asset_type == "toolbit":
                    # Look for toolbit files in parallel Bit directory
                    # Library is in Library/, toolbits are in parallel Bit/
                    library_dir = file_path.parent  # e.g., /path/to/Library/
                    tools_dir = library_dir.parent  # e.g., /path/to/
                    bit_dir = tools_dir / "Bit"  # e.g., /path/to/Bit/

                    if bit_dir.exists():
                        possible_extensions = [".fctb", ".json", ".yaml", ".yml"]
                        for ext in possible_extensions:
                            toolbit_file = bit_dir / f"{dependency_uri.asset_id}{ext}"
                            if toolbit_file.exists():
                                dependency_found = True
                                external_toolbits.append((dependency_uri, toolbit_file))
                                break

                if not dependency_found:
                    message = f"Failed to import {file_path}: required dependency {dependency_uri} not found in stores or in parallel Bit directory"
                    if quiet:
                        Path.Log.error(message)
                    else:
                        QMessageBox.critical(
                            self,
                            "Error",
                            message,
                        )
                    return None

            # If we found external toolbits, ask user if they want to import them
            if external_toolbits:
                toolbit_names = [uri.asset_id for uri, _ in external_toolbits]
                if quiet:
                    Path.Log.info("Importing tool bits for the library")
                    reply = QMessageBox.Yes
                else:
                    reply = QMessageBox.question(
                        self,
                        "Import External Toolbits",
                        f"This library references {len(external_toolbits)} toolbit(s) that are not in your local store:\n\n"
                        + "\n".join(f"• {name}" for name in toolbit_names)
                        + f"\n\nWould you like to import these toolbits into your local store?\n"
                        + "This will make them permanently available for use in other libraries.",
                        QMessageBox.Yes | QMessageBox.No,
                        QMessageBox.Yes,
                    )

                if reply == QMessageBox.Yes:
                    # Import the external toolbits into local store
                    self._import_external_toolbits(external_toolbits, quiet=quiet)
                    # After importing, use regular deserialization since toolbits are now in local store
                else:
                    # User declined import, use context-aware deserialization for external loading
                    pass
            else:
                # No external toolbits found, use regular deserialization
                pass
        except Exception as e:
            QMessageBox.critical(self, "Error", f"{file_path}: Failed to check dependencies: {e}")
            return None

        # Load and return the asset.
        try:
            # Always use context-aware deserialization for libraries to get meaningful names
            if hasattr(serializer_class, "deep_deserialize_with_context"):
                # Pass file path context for meaningful library names and external dependency resolution
                asset = serializer_class.deep_deserialize_with_context(raw_data, file_path)
            else:
                # Fallback to regular deserialization
                asset = serializer_class.deep_deserialize(raw_data)

            if not isinstance(asset, self.asset_class):
                raise TypeError(f"Deserialized asset is not of type {self.asset_class.__name__}")
            return asset
        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to import asset: {e}")
            return None

    def exec_(self) -> Optional[Tuple[pathlib.Path, Asset]]:
        if super().exec_():
            filenames = self.selectedFiles()
            if filenames:
                file_path = pathlib.Path(filenames[0])
                asset = self.deserialize_file(file_path)
                if asset:
                    return file_path, asset
        return None

    def _get_default_directory(self, asset_class: Type[Asset]) -> pathlib.Path:
        """Get the appropriate default directory based on asset type."""
        try:
            asset_path = Preferences.getAssetPath()

            # Check asset type to determine subdirectory
            asset_type = getattr(asset_class, "asset_type", None)
            if asset_type == "toolbit":
                return asset_path / "Tool" / "Bit"
            elif asset_type == "library" or asset_type == "toolbitlibrary":
                return asset_path / "Tool" / "Library"
            else:
                # Default to asset path root for unknown types
                return asset_path
        except Exception:
            # Fallback to home directory if anything goes wrong
            return pathlib.Path.home()

    def _import_external_toolbits(self, external_toolbits, quiet=False):
        """Import external toolbits into the local asset store."""
        from ...toolbit.serializers import all_serializers as toolbit_serializers
        from .util import get_serializer_from_extension

        imported_count = 0
        failed_imports = []

        for dependency_uri, toolbit_file in external_toolbits:
            try:
                # Find appropriate serializer for the file
                file_extension = toolbit_file.suffix.lower()
                serializer_class = get_serializer_from_extension(
                    toolbit_serializers, file_extension, for_import=True
                )

                if not serializer_class:
                    failed_imports.append(
                        f"{dependency_uri.asset_id}: No serializer for {file_extension}"
                    )
                    continue

                # Load and deserialize the toolbit
                raw_toolbit_data = toolbit_file.read_bytes()
                toolbit = serializer_class.deep_deserialize(raw_toolbit_data)

                # Ensure the toolbit ID matches what the library expects
                if toolbit.id != dependency_uri.asset_id:
                    toolbit.id = dependency_uri.asset_id

                # Import the toolbit into local store
                cam_assets.add(toolbit)
                imported_count += 1

            except Exception as e:
                failed_imports.append(f"{dependency_uri.asset_id}: {str(e)}")

        # Show results to user
        if imported_count > 0:
            message = f"Successfully imported {imported_count} toolbit(s) into your local store."
            if failed_imports:
                message += f"\n\nFailed to import {len(failed_imports)} toolbit(s):\n" + "\n".join(
                    failed_imports
                )
            if quiet:
                Path.Log.info(message)
            else:
                QMessageBox.information(self, "Import Results", message)
        elif failed_imports:
            message = f"Failed to import all {len(failed_imports)} toolbit(s):\n" + "\n".join(
                failed_imports
            )
            if quiet:
                Path.Log.error(message)
            else:
                QMessageBox.warning(self, "Import Failed", message)


class AssetSaveDialog(QFileDialog):
    def __init__(
        self,
        asset_class: Type[Asset],
        serializers: Iterable[Type[AssetSerializer]],
        parent=None,
    ):
        super().__init__(parent)

        # Set default directory based on asset type
        default_dir = self._get_default_directory(asset_class)
        self.setDirectory(default_dir.as_posix())
        self.asset_class = asset_class
        self.serializers = list(serializers)
        self.setFileMode(QFileDialog.AnyFile)
        self.setAcceptMode(QFileDialog.AcceptSave)
        self.filters, self.serializer_map = make_export_filters(self.serializers)
        self.setNameFilters(self.filters)
        if self.filters:
            self.selectNameFilter(self.filters[0])  # Default to "Automatic"
        self.filterSelected.connect(self.update_default_suffix)

    def update_default_suffix(self, filter_str: str):
        """Update the default suffix based on the selected filter."""
        if filter_str == "Automatic (*)":
            self.setDefaultSuffix("")  # No default for Automatic
        else:
            serializer = self.serializer_map.get(filter_str)
            if serializer and serializer.extensions:
                self.setDefaultSuffix(serializer.extensions[0])

    def _serialize_selected_file(
        self,
        file_path: pathlib.Path,
        asset: Asset,
        serializer_class: Type[AssetSerializer],
    ) -> bool:
        """Serialize and save the asset."""
        try:
            raw_data = serializer_class.serialize(asset)
            file_path.write_bytes(raw_data)
            return True
        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to export asset: {e}")
            return False

    def _get_default_directory(self, asset_class: Type[Asset]) -> pathlib.Path:
        """Get the appropriate default directory based on asset type."""
        # For exports, default to home directory instead of CAM assets path
        return pathlib.Path.home()

    def exec_(self, asset: Asset) -> Optional[Tuple[pathlib.Path, Type[AssetSerializer]]]:
        self.setWindowTitle(f"Save {asset.label or self.asset_class.asset_type}")
        if super().exec_():
            selected_filter = self.selectedNameFilter()
            file_path = pathlib.Path(self.selectedFiles()[0])
            if selected_filter == "Automatic (*)":
                if not file_path.suffix:
                    QMessageBox.critical(
                        self,
                        "Error",
                        "Please specify a file extension for automatic serializer selection.",
                    )
                    return None
                file_extension = file_path.suffix.lower()
                serializer_class = get_serializer_from_extension(
                    self.serializers, file_extension, for_import=False
                )
                if not serializer_class:
                    QMessageBox.critical(
                        self,
                        "Error",
                        f"No serializer found for extension '{file_extension}'",
                    )
                    return None
            else:
                serializer_class = self.serializer_map.get(selected_filter)
                if not serializer_class:
                    raise ValueError(f"No serializer found for filter '{selected_filter}'")
            if self._serialize_selected_file(file_path, asset, serializer_class):
                return file_path, serializer_class
        return None



# SPDX-License-Identifier: LGPL-2.1-or-later

################################################################################
#                                                                              #
#   © 2026 Billy Huddleston <billy@ivdc.com>                                   #
#                                                                              #
#   FreeCAD is free software: you can redistribute it and/or modify            #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   FreeCAD is distributed in the hope that it will be useful,                 #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with FreeCAD. If not, see https://www.gnu.org/licenses       #
#                                                                              #
################################################################################
import FreeCAD
import Path
from ...shape import ToolBitShapeTaperedBallNose
from ..mixins import RotaryToolBitMixin, CuttingToolMixin
from .base import ToolBit


class ToolBitTaperedBallNose(ToolBit, CuttingToolMixin, RotaryToolBitMixin):
    SHAPE_CLASS = ToolBitShapeTaperedBallNose

    def __init__(self, shape: ToolBitShapeTaperedBallNose, id: str | None = None):
        Path.Log.track(f"ToolBitTaperedBallNose __init__ called with shape: {shape}, id: {id}")
        super().__init__(shape, id=id)
        CuttingToolMixin.__init__(self, self.obj)

    @property
    def summary(self) -> str:
        diameter = self.get_property_str("Diameter", "?", precision=3)
        flutes = self.get_property("Flutes")
        cutting_edge_height = self.get_property_str("CuttingEdgeHeight", "?", precision=3)
        taper_angle = self.get_property_str("TaperAngle", "?", precision=1)

        return FreeCAD.Qt.translate(
            "CAM",
            f"{diameter} tip, {taper_angle} taper, {flutes}-flute tapered ball nose, {cutting_edge_height} cutting edge",
        )



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint centrif"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


import math

import FreeCAD


def get_analysis_types():
    return ["buckling", "static", "thermomech"]


def get_sets_name():
    return "constraints_centrif_element_sets"


def get_constraint_title():
    return "Centrif Constraints"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_meshdata_constraint(f, femobj, centrif_obj, ccxwriter):

    f.write(f"*ELSET,ELSET={centrif_obj.Name}\n")
    for refs, elem, is_sub_el in femobj["CentrifElements"]:
        if not is_sub_el:
            for e in elem:
                f.write(f"{e},\n")

    # use all elements if there are no references
    if not femobj["CentrifElements"]:
        f.write(f"{ccxwriter.ccx_eall}\n")


def write_constraint(f, femobj, centrif_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    # get some data from the centrif_obj
    refobj = centrif_obj.RotationAxis[0][0]
    subobj = centrif_obj.RotationAxis[0][1][0]
    axis = refobj.Shape.getElement(subobj)

    if axis.Curve.TypeId == "Part::GeomLine":
        axiscopy = axis.copy()  # apply global placement to copy
        axiscopy.Placement = refobj.getGlobalPlacement()
        direction = axiscopy.Curve.Direction
        location = axiscopy.Curve.Location
    else:  # no line found, set default
        # TODO: No test at all in the writer
        # they should all be before in prechecks
        location = FreeCAD.Vector(0.0, 0.0, 0.0)
        direction = FreeCAD.Vector(0.0, 0.0, 1.0)

    # write to file
    f.write("*DLOAD\n")
    f.write(
        "{},CENTRIF,{:.13G},{:.13G},{:.13G},{:.13G},{:.13G},{:.13G},{:.13G}\n".format(
            centrif_obj.Name,
            (2.0 * math.pi * float(centrif_obj.RotationFrequency.getValueAs("1/s"))) ** 2,
            location.x,
            location.y,
            location.z,
            direction.x,
            direction.y,
            direction.z,
        )
    )
    f.write("\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD task panel for CalculiX solver"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package task_solver_calculix
#  \ingroup FEM
#  \brief task panel for CalculiX solver

from PySide import QtCore
from PySide import QtGui

import FreeCAD
import FreeCADGui

import FemGui

from femsolver.calculix import calculixtools

from . import base_femlogtaskpanel


class _TaskPanel(base_femlogtaskpanel._BaseLogTaskPanel):
    """
    The TaskPanel for run CalculiX solver
    """

    def __init__(self, obj):
        super().__init__(obj, calculixtools.CalculiXTools(obj))

        self.form = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/SolverCalculiX.ui"
        )

        self.text_log = self.form.te_output
        self.text_time = self.form.l_time
        self.prepared = False
        self.run_complete = False

        self.setup_connections()

    def setup_connections(self):
        super().setup_connections()

        QtCore.QObject.connect(
            self.form.ckb_working_directory,
            QtCore.SIGNAL("toggled(bool)"),
            self.working_directory_toggled,
        )
        QtCore.QObject.connect(
            self.form.cb_analysis_type,
            QtCore.SIGNAL("currentIndexChanged(int)"),
            self.analysis_type_changed,
        )
        QtCore.QObject.connect(
            self.form.pb_write_input, QtCore.SIGNAL("clicked()"), self.write_input_clicked
        )
        QtCore.QObject.connect(
            self.form.pb_edit_input, QtCore.SIGNAL("clicked()"), self.edit_input_clicked
        )
        QtCore.QObject.connect(
            self.form.fc_working_directory,
            QtCore.SIGNAL("fileNameSelected(QString)"),
            self.working_directory_selected,
        )
        QtCore.QObject.connect(
            self.form.pb_solver_version, QtCore.SIGNAL("clicked()"), self.get_version
        )

        self.get_object_params()
        self.set_widgets()

    def preparation_finished(self):
        # override base class method to not auto compute
        self.prepared = True
        if not self.run_complete:
            self.timer.stop()
            self.form.pb_edit_input.setEnabled(True)
        else:
            super().preparation_finished()

    def apply(self):
        self.text_log.clear()
        self.elapsed.restart()
        if self.prepared:
            self.timer.start(100)
            self.tool.compute()
        else:
            # run complete process if 'Apply' is pressed without
            # previously write the input files
            self.run_complete = True
            super().apply()

    def get_object_params(self):
        self.analysis_type = self.obj.AnalysisType

    def set_object_params(self):
        self.obj.AnalysisType = self.analysis_type

    def set_widgets(self):
        "fills the widgets"
        self.analysis_type_enum = self.obj.getEnumerationsOfProperty("AnalysisType")
        index = self.analysis_type_enum.index(self.analysis_type)
        self.form.cb_analysis_type.addItems(self.analysis_type_enum)
        self.form.cb_analysis_type.setCurrentIndex(index)

        self.form.fc_working_directory.setProperty("fileName", self.obj.WorkingDirectory)
        self.form.ckb_working_directory.setChecked(False)
        self.form.gpb_working_directory.setVisible(False)

    def analysis_type_changed(self, index):
        self.analysis_type = self.analysis_type_enum[index]
        self.obj.AnalysisType = self.analysis_type

    def working_directory_selected(self):
        self.obj.WorkingDirectory = self.form.fc_working_directory.property("fileName")

    def write_input_clicked(self):
        self.prepared = False
        self.run_complete = False
        self.run_process()

    def edit_input_clicked(self):
        ccx_param = self.tool.fem_param.GetGroup("Ccx")
        internal = ccx_param.GetBool("UseInternalEditor", True)
        ext_editor_path = ccx_param.GetString("ExternalEditorPath", "")
        if internal or not ext_editor_path:
            FemGui.open(self.tool.model_file)
        else:
            ext_editor_process = QtCore.QProcess()
            ext_editor_process.start(ext_editor_path, [self.tool.model_file])
            ext_editor_process.waitForFinished()

    def working_directory_toggled(self, bool_value):
        self.form.gpb_working_directory.setVisible(bool_value)



# ***************************************************************************
# *   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import unittest
import math

import FreeCAD
from FreeCAD import Base
import Part
import Sketcher


class TestShapeBinder(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestShapeBinder")

    def testTwoBodyShapeBinderCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Box = self.Doc.addObject("PartDesign::AdditiveBox", "Box")
        self.Box.Length = 1
        self.Box.Width = 1
        self.Box.Height = 1
        self.Body.addObject(self.Box)
        self.Doc.recompute()
        self.Body001 = self.Doc.addObject("PartDesign::Body", "Body001")
        self.ShapeBinder = self.Doc.addObject("PartDesign::ShapeBinder", "ShapeBinder")
        self.ShapeBinder.Support = [(self.Box, "Face1")]
        self.Body001.addObject(self.ShapeBinder)
        self.Doc.recompute()
        self.assertIn("Box", self.ShapeBinder.OutList[0].Label)
        self.assertIn("Body001", self.ShapeBinder.InList[0].Label)

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestShapeBinder")
        # print ("omit closing document for debugging")


class TestSubShapeBinder(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestSubShapeBinder")

    def tearDown(self):
        FreeCAD.closeDocument("PartDesignTestSubShapeBinder")

    def testOffsetBinder(self):
        # See PR 7445
        body = self.Doc.addObject("PartDesign::Body", "Body")
        box = self.Doc.addObject("PartDesign::AdditiveBox", "Box")
        body.addObject(box)

        box.Length = 10.00000
        box.Width = 10.00000
        box.Height = 10.00000

        binder = body.newObject("PartDesign::SubShapeBinder", "Binder")
        binder.Support = [(box, ("Edge2", "Edge12", "Edge6", "Edge10"))]
        self.Doc.recompute()

        self.assertAlmostEqual(binder.Shape.Length, 40)

        binder.OffsetJoinType = "Tangent"
        binder.Offset = 5.00000
        self.Doc.recompute()

        self.assertAlmostEqual(binder.Shape.Length, 80)

    def testBinderBeforeOrAfterPad(self):
        """Test case for PR #8763"""
        body = self.Doc.addObject("PartDesign::Body", "Body")
        sketch = body.newObject("Sketcher::SketchObject", "Sketch")
        sketch.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        sketch.MapMode = "FlatFace"
        self.Doc.recompute()

        geoList = []
        geoList.append(
            Part.LineSegment(
                Base.Vector(-21.762587, 19.904083, 0), Base.Vector(32.074337, 19.904083, 0)
            )
        )
        geoList.append(
            Part.LineSegment(
                Base.Vector(32.074337, 19.904083, 0), Base.Vector(32.074337, -27.458027, 0)
            )
        )
        geoList.append(
            Part.LineSegment(
                Base.Vector(32.074337, -27.458027, 0), Base.Vector(-21.762587, -27.458027, 0)
            )
        )
        geoList.append(
            Part.LineSegment(
                Base.Vector(-21.762587, -27.458027, 0), Base.Vector(-21.762587, 19.904083, 0)
            )
        )
        sketch.addGeometry(geoList, False)

        conList = []
        conList.append(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
        conList.append(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
        conList.append(Sketcher.Constraint("Coincident", 2, 2, 3, 1))
        conList.append(Sketcher.Constraint("Coincident", 3, 2, 0, 1))
        conList.append(Sketcher.Constraint("Horizontal", 0))
        conList.append(Sketcher.Constraint("Horizontal", 2))
        conList.append(Sketcher.Constraint("Vertical", 1))
        conList.append(Sketcher.Constraint("Vertical", 3))
        sketch.addConstraint(conList)
        del geoList, conList

        self.Doc.recompute()

        binder1 = body.newObject("PartDesign::SubShapeBinder", "Binder")
        binder1.Support = sketch
        self.Doc.recompute()
        pad = body.newObject("PartDesign::Pad", "Pad")
        pad.Profile = sketch
        pad.Length = 10
        self.Doc.recompute()
        pad.ReferenceAxis = (sketch, ["N_Axis"])
        sketch.Visibility = False
        self.Doc.recompute()

        binder2 = body.newObject("PartDesign::SubShapeBinder", "Binder001")
        binder2.Support = [pad, "Sketch."]
        self.Doc.recompute()

        self.assertAlmostEqual(binder1.Shape.BoundBox.XLength, binder2.Shape.BoundBox.XLength, 2)
        self.assertAlmostEqual(binder1.Shape.BoundBox.YLength, binder2.Shape.BoundBox.YLength, 2)
        self.assertAlmostEqual(binder1.Shape.BoundBox.ZLength, binder2.Shape.BoundBox.ZLength, 2)

        nor1 = binder1.Shape.Face1.normalAt(0, 0)
        nor2 = binder2.Shape.Face1.normalAt(0, 0)
        self.assertAlmostEqual(nor1.getAngle(nor2), 0.0, 2)

    def testBinderWithRevolution(self):
        doc = self.Doc
        body = doc.addObject("PartDesign::Body", "Body")
        doc.recompute()
        sketch = body.newObject("Sketcher::SketchObject", "Sketch")
        geoList = []
        geoList.append(Part.LineSegment(Base.Vector(10, 10, 0), Base.Vector(30, 10, 0)))
        geoList.append(Part.LineSegment(Base.Vector(30, 10, 0), Base.Vector(30, 15, 0)))
        geoList.append(Part.LineSegment(Base.Vector(30, 15, 0), Base.Vector(10, 15, 0)))
        geoList.append(Part.LineSegment(Base.Vector(10, 15, 0), Base.Vector(10, 10, 0)))
        sketch.addGeometry(geoList, False)
        del geoList
        constraintList = []
        constraintList.append(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
        constraintList.append(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
        constraintList.append(Sketcher.Constraint("Coincident", 2, 2, 3, 1))
        constraintList.append(Sketcher.Constraint("Coincident", 3, 2, 0, 1))
        constraintList.append(Sketcher.Constraint("Horizontal", 0))
        constraintList.append(Sketcher.Constraint("Horizontal", 2))
        constraintList.append(Sketcher.Constraint("Vertical", 1))
        constraintList.append(Sketcher.Constraint("Vertical", 3))
        sketch.addConstraint(constraintList)
        del constraintList
        doc.recompute()
        binder = body.newObject("PartDesign::SubShapeBinder", "Binder")
        binder.Support = sketch
        revolution = body.newObject("PartDesign::Revolution", "Revolution")
        revolution.Profile = (
            binder,
            [
                "",
            ],
        )
        revolution.ReferenceAxis = (doc.getObject("Y_Axis"), [""])
        revolution.Angle = 360.0
        revolution.Reversed = 1
        doc.recompute()
        revolution.Angle2 = 60.0
        doc.recompute()
        self.assertAlmostEqual(binder.Shape.Area, 100)
        volume = 100 * math.pi * 2 * 20
        self.assertAlmostEqual(revolution.Shape.Volume, volume)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 baidakovil <baidakovil@icloud.com>                 *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************


"""
Unit tests for WebGL export functionality. Gui tests are in `TestWebGLExportGui.py`.
"""

import os
import tempfile
import unittest
from unittest.mock import patch, mock_open

from BIM.importers import importWebGL
from .TestArchBase import TestArchBase


class TestWebGLExport(TestArchBase):

    def setUp(self):
        """Using TestArchBase setUp to initialize the document for convenience,
        but also create a temporary directory for tests."""
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.test_template_content = """
        <!DOCTYPE html>
        <html>
        <head><title>$pagetitle</title></head>
        <body>WebGL Content: $data</body>
        </html>
        """

    def tearDown(self):
        import shutil

        shutil.rmtree(self.test_dir, ignore_errors=True)
        super().tearDown()

    def test_actual_default_template_readable_and_valid(self):
        """Test that the actual default template can be read and contains
        required placeholders"""
        operation = "Testing actual default template reading and validation"
        self.printTestMessage(operation)
        # Test with real configuration - no mocks
        with patch(
            "BIM.importers.importWebGL.params.get_param", return_value=False
        ):  # Disable custom template
            result = importWebGL.getHTMLTemplate()

            if result is not None:  # Only test if template exists
                self.assertIsInstance(result, str)

                # Check for basic HTML structure
                self.assertIn("<html", result.lower())
                self.assertIn("$pagetitle", result.lower())
                self.assertIn("$data", result.lower())

    def test_custom_template_with_real_file(self):
        """Test custom template functionality with real temporary file"""
        operation = "Testing custom template with real file"
        self.printTestMessage(operation)
        # Create real temporary template file
        custom_path = os.path.join(self.test_dir, "custom_template.html")
        with open(custom_path, "w", encoding="utf-8") as f:
            f.write(self.test_template_content)

        with patch("BIM.importers.importWebGL.params.get_param") as mock_params:
            mock_params.side_effect = lambda param, path=None: {
                "useCustomWebGLExportTemplate": True,
                "WebGLTemplateCustomPath": custom_path,
            }.get(param, False)

            result = importWebGL.getHTMLTemplate()
            self.assertIsNotNone(result)
            self.assertEqual(result.strip(), self.test_template_content.strip())

    def test_default_template_logic_when_custom_disabled(self):
        """Test code path when custom template is disabled"""
        operation = "Testing default template logic when custom is disabled"
        self.printTestMessage(operation)
        with patch("BIM.importers.importWebGL.params.get_param", return_value=False):
            with patch("os.path.isfile", return_value=True):
                with patch(
                    "builtins.open",
                    mock_open(read_data=self.test_template_content),
                ):
                    result = importWebGL.getHTMLTemplate()
                    self.assertIsNotNone(result)
                    if result is not None:
                        self.assertIn("WebGL Content", result)

    def test_custom_template_not_found_headless_mode(self):
        """Test behavior when custom template not found in headless mode"""
        operation = "Testing custom template not found in headless mode"
        self.printTestMessage(operation)
        with patch("BIM.importers.importWebGL.params.get_param") as mock_params:
            mock_params.side_effect = lambda param, path=None: {
                "useCustomWebGLExportTemplate": True,
                "WebGLTemplateCustomPath": "/nonexistent/template.html",
            }.get(param, False)

            with patch("BIM.importers.importWebGL.FreeCADGui", None):  # Simulate headless mode
                result = importWebGL.getHTMLTemplate()
                self.assertIsNone(result)


if __name__ == "__main__":
    # Allow running tests directly
    unittest.main(verbosity=2)



# SPDX-License-Identifier: LGPL-2.1-or-later

from .editor import ToolBitEditorPanel, ToolBitEditor
from .browser import ToolBitBrowserWidget

__all__ = [
    "ToolBitBrowserWidget",
    "ToolBitEditor",
    "ToolBitEditorPanel",
]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import FreeCAD as App
import Arch
import Draft
from bimtests import TestArchBase


class TestArchFence(TestArchBase.TestArchBase):

    def test_makeFence(self):
        # Create section, post, and path objects
        section = Draft.makeRectangle(100, 10)
        post = Draft.makeRectangle(10, 10)
        path = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(1000, 0, 0))

        # Create a fence
        fence = Arch.makeFence(section, post, path)
        self.assertIsNotNone(fence, "Failed to create fence")
        self.assertEqual(fence.Section, section, "Fence section is incorrect")
        self.assertEqual(fence.Post, post, "Fence post is incorrect")
        self.assertEqual(fence.Path, path, "Fence path is incorrect")



# ***************************************************************************
# *   Copyright (c) 2022 Wanderer Fan <wandererfan@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the TechDraw ShareView Task Dialog."""

__title__ = "TechDrawTools.TaskShareView"
__author__ = "WandererFan"
__url__ = "https://www.freecad.org"
__version__ = "00.01"
__date__ = "2022/01/11"

from PySide import QtCore
import PySide.QtGui as QtGui

import FreeCAD as App
import FreeCADGui as Gui

from TechDrawTools import TDToolsMovers

import os

translate = App.Qt.translate

class TaskShareView:
    def __init__(self):
        self._uiPath = App.getHomePath()
        self._uiPath = os.path.join(self._uiPath, "Mod/TechDraw/TechDrawTools/Gui/TaskMoveView.ui")
        self.form = Gui.PySideUic.loadUi(self._uiPath)

        self.form.setWindowTitle(translate("TechDraw_ShareView", "Share View With Another Page"))
        self.form.lViewName.setText(translate("TechDraw_ShareView", "View to share"))

        self.form.pbView.clicked.connect(self.pickView)
        self.form.pbFromPage.clicked.connect(self.pickFromPage)
        self.form.pbToPage.clicked.connect(self.pickToPage)

        self.viewName = ""
        self.fromPageName = ""
        self.toPageName   = ""

        self.dialogOpen = False

        App.setActiveTransaction("Share view")

    def accept(self):
#        print ("Accept")
        view = App.ActiveDocument.getObject(self.viewName)
        fromPage = App.ActiveDocument.getObject(self.fromPageName)
        toPage = App.ActiveDocument.getObject(self.toPageName)
        TDToolsMovers.moveView(view, fromPage, toPage, True)

        App.closeActiveTransaction()
        return True

    def reject(self):
#        print ("Reject")
        App.closeActiveTransaction(True)
        return True

    def pickView(self):
#        print("pickView")
        if (self.dialogOpen) :
            return
        _dlgPath = App.getHomePath()
        _dlgPath = os.path.join(_dlgPath, "Mod/TechDraw/TechDrawTools/Gui/DlgPageChooser.ui")
        dlg = Gui.PySideUic.loadUi(_dlgPath)
        self.dialogOpen = True
        dlg.lPrompt.setText(translate("TechDraw_ShareView", "Select view to share from list."))
        dlg.setWindowTitle(translate("TechDraw_ShareView", "Select View"))

        views = [x for x in App.ActiveDocument.Objects if x.isDerivedFrom("TechDraw::DrawView")]
        for v in views:
            s = v.Label + " / " + v.Name
            item = QtGui.QListWidgetItem(s, dlg.lwPages)
            item.setData(QtCore.Qt.UserRole, v.Name)
        if (dlg.exec() == QtGui.QDialog.Accepted) :
            if dlg.lwPages.selectedItems():
                selItem = dlg.lwPages.selectedItems()[0]
                self.viewName = selItem.data(QtCore.Qt.UserRole)
                self.form.leView.setText(self.viewName)
        self.dialogOpen = False

    def pickFromPage(self):
#        print("pickFromPage")
        if (self.dialogOpen) :
            return
        _dlgPath = App.getHomePath()
        _dlgPath = os.path.join(_dlgPath, "Mod/TechDraw/TechDrawTools/Gui/DlgPageChooser.ui")
        dlg = Gui.PySideUic.loadUi(_dlgPath)
        self.dialogOpen = True
        dlg.lPrompt.setText(translate("TechDraw_ShareView", "Select from page."))
        dlg.setWindowTitle(translate("TechDraw_ShareView", "Select Page"))

        pages = [x for x in App.ActiveDocument.Objects if x.isDerivedFrom("TechDraw::DrawPage")]
        for p in pages:
            s = p.Label + " / " + p.Name
            item = QtGui.QListWidgetItem(s, dlg.lwPages)
            item.setData(QtCore.Qt.UserRole, p.Name)
        if (dlg.exec() == QtGui.QDialog.Accepted) :
            if dlg.lwPages.selectedItems():
                selItem = dlg.lwPages.selectedItems()[0]
                self.fromPageName = selItem.data(QtCore.Qt.UserRole)
                self.form.leFromPage.setText(self.fromPageName)
        self.dialogOpen = False


    def pickToPage(self):
#        print("pickToPage")
        if (self.dialogOpen) :
            return
        _dlgPath = App.getHomePath()
        _dlgPath = os.path.join(_dlgPath, "Mod/TechDraw/TechDrawTools/Gui/DlgPageChooser.ui")
        dlg = Gui.PySideUic.loadUi(_dlgPath)
        self.dialogOpen = True
        dlg.lPrompt.setText(translate("TechDraw_ShareView", "Select to page."))
        dlg.setWindowTitle(translate("TechDraw_ShareView", "Select Page"))

        pages = [x for x in App.ActiveDocument.Objects if x.isDerivedFrom("TechDraw::DrawPage")]
        for p in pages:
            s = p.Label + " / " + p.Name
            item = QtGui.QListWidgetItem(s, dlg.lwPages)
            item.setData(QtCore.Qt.UserRole, p.Name)
        if (dlg.exec() == QtGui.QDialog.Accepted) :
            if dlg.lwPages.selectedItems():
                selItem = dlg.lwPages.selectedItems()[0]
                self.toPageName = selItem.data(QtCore.Qt.UserRole)
                self.form.leToPage.setText(self.toPageName)
        self.dialogOpen = False

    def setValues(self, viewName, fromPageName, toPageName):
        self.viewName = viewName
        self.form.leView.setText(viewName)
        self.fromPageName = fromPageName
        self.form.leFromPage.setText(fromPageName)
        self.toPageName = toPageName
        self.form.leToPage.setText(toPageName)




# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Ondsel <development@ondsel.com>                    *
# *   Copyright (c) 2025 Billy Huddleston <billy@ivdc.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
from string import Template

"""
This module contains the HTML template for the CAM Sanity report.
"""

html_template = Template(
    """
<!DOCTYPE html>
<html>
<head>
    <meta http-equiv="content-type" content="text/html; charset=utf-8"/>
    <title>Setup Report for FreeCAD Job: Path Special</title>
    <style type="text/css">
        /* Reset margins and padding */
        div, dl, dt, dd, ul, ol, li, h1, h2, h3, #toctitle, .sidebarblock>.content>.title, h4, h5, h6, pre, form, p, blockquote, th, td {
            margin: 0;
            padding: 0;
        }

        /* Responsive image handling */
        img, object, embed {
            max-width: 100%;
            height: auto;
        }

        /* Base styling */
        body {
            background: #fff;
            color: rgba(0,0,0,.8);
            padding: 0;
            margin: 0;
            font-family: "Noto Serif", "DejaVu Serif", serif;
            line-height: 1;
            position: relative;
            cursor: auto;
            word-wrap: anywhere;
            -moz-osx-font-smoothing: grayscale;
            -webkit-font-smoothing: antialiased;
            max-width: 62.5em;
            margin-left: auto;
            margin-right: auto;
            padding: 20px;
        }

        /* Container styling */
        .center-container {
            max-width: 62.5em;
            margin-left: auto;
            margin-right: auto;
            padding: 20px;
        }

        /* Headings */
        h1, h2, h3, #toctitle, h4, h5, h6 {
            font-family: "Open Sans", "DejaVu Sans", sans-serif;
            font-weight: 300;
            font-style: normal;
            color: #ba3925;
            text-rendering: optimizeLegibility;
            margin-top: 1em;
            margin-bottom: .5em;
            line-height: 1.0125em;
        }

        h1 {
            font-size: 2.125em;
        }

        h2 {
            font-size: 1.6875em;
        }

        h3 {
            font-size: 1.375em;
        }

        h4, h5 {
            font-size: 1.125em;
        }

        h6 {
            font-size: 1em;
        }

        #toctitle {
            font-size: 1.2em;
        }

        @media screen and (min-width: 768px) {
            #toctitle {
                font-size: 1.375em;
            }
        }

        /* Media query for larger screens */
        @media screen and (min-width: 768px) {
            h1, h2, h3, h4, h5, h6 {
                line-height: 1.2;
            }

            h1 {
                font-size: 2.75em;
            }

            h2 {
                font-size: 2.3125em;
            }

            h3 {
                font-size: 1.6875em;
            }

            h4 {
                font-size: 1.4375em;
            }
        }

        /* Links */
        a {
            color: #2156a5;
            text-decoration: underline;
            line-height: inherit;
        }

        /* TOC styling */
        #toc {
            border-bottom: 1px solid #e7e7e9;
            padding-bottom: .5em;
        }

        #header>h1:first-child+#toc {
            margin-top: 8px;
            border-top: 1px solid #dddddf;
        }

        #toc>ul {
            margin-left: .125em;
        }

        #toc ul {
            font-family: "Open Sans", "DejaVu Sans", sans-serif;
            list-style-type: none;
        }

        #toc li {
            line-height: 1.3334;
            margin-top: .3334em;
        }

        #toc a {
            text-decoration: none;
        }

        a:hover, a:focus {
            color: #1d4b8f;
        }

        /* Lists */
        ul, ol, dl {
            line-height: 1.6;
            margin-bottom: 1.25em;
            list-style-position: outside;
            font-family: inherit;
            margin-left: 1.5em;
        }

        ul {
            list-style: disc;
        }

        ul li ul, ul li ol {
            margin-left: 1.25em;
            margin-bottom: 0;
        }

        /* TOC section levels */
        #toc ul.sectlevel0>li>a {
            font-style: italic;
        }

        #toc ul.sectlevel0 ul.sectlevel1 {
            margin: .5em 0;
        }

        @media screen and (min-width: 768px) {
            #toc.toc2 ul ul {
                margin-left: 0;
                padding-left: 1em;
            }

            #toc.toc2 ul.sectlevel0 ul.sectlevel1 {
                padding-left: 0;
                margin-top: .5em;
                margin-bottom: .5em;
            }
        }

        @media screen and (min-width: 1280px) {
            #toc.toc2 ul ul {
                padding-left: 1.25em;
            }
        }

        /* Responsive image handling at different screen sizes */
        @media screen and (max-width: 768px) {
            td.image-container {
                max-width: 100%;
                display: block;
                margin: 1em auto;
            }

            table td[rowspan] {
                display: table-cell;
            }

            img {
                max-width: 100%;
                height: auto;
            }
        }

        /* Tables - more detailed styling */
        table {
            background: #fff;
            margin-bottom: 1.25em;
            border: 1px solid #dedede;
            word-wrap: normal;
            border-collapse: collapse;
            border-spacing: 0;
            width: 100%;
        }

        table thead, table tfoot {
            background: #f7f8f7;
        }

        table thead tr th, table thead tr td, table tfoot tr th, table tfoot tr td {
            padding: .5em .625em .625em;
            font-size: inherit;
            color: rgba(0,0,0,.8);
            text-align: left;
        }

        table tr th, table tr td {
            padding: .5625em .625em;
            font-size: inherit;
            color: rgba(0,0,0,.8);
            line-height: 1.6;
            border: 1px solid #dedede;
        }

        /* Anchor links styling */
        #content h1>a.anchor,h2>a.anchor,h3>a.anchor,#toctitle>a.anchor,.sidebarblock>.content>.title>a.anchor,h4>a.anchor,h5>a.anchor,h6>a.anchor{
            position:absolute;
            z-index:1001;
            width:1.5ex;
            margin-left:-1.5ex;
            display:block;
            text-decoration:none!important;
            visibility:hidden;
            text-align:center;
            font-weight:400
        }

        table tr.even, table tr.alt {
            background: #f8f8f7;
        }

        /* Image container styling */
        td.image-container {
            vertical-align: middle;
            text-align: center;
            width: 40%;
            max-width: 300px;
        }

        td.image-container img {
            max-width: 100%;
            height: auto;
            object-fit: contain;
        }

        /* Text styling */
        p {
            line-height: 1.6;
            margin-bottom: 1.25em;
            text-rendering: optimizeLegibility;
        }

        strong, b {
            font-weight: bold;
            line-height: inherit;
        }

        em, i {
            font-style: italic;
            line-height: inherit;
        }

        /* Top navigation links - base style */
        .top-link {
            display: inline-block;
            float: right;
            font-size: 0.8em;
            font-weight: normal;
            text-transform: uppercase;
            color: #2156a5;
            text-decoration: none;
            position: relative;
        }

        /* Specific positioning for h2 headings */
        .heading-container h2 + .top-link {
            margin-top: 4.75em;
            bottom: 0.3em;
        }

        /* Specific positioning for h3 headings */
        .heading-container h3 + .top-link {
            margin-top: 3.5em;
            bottom: 0.3em;
        }

        /* Fallbacks in case the adjacent sibling selector doesn't work as expected */
        .heading-container:has(h2) .top-link {
            margin-top: 4.75em;
        }

        .heading-container:has(h3) .top-link {
            margin-top: 3.5em;
        }

        /* Clearfix for headings with top links */
        .heading-container {
            overflow: hidden;
            width: 100%;
            position: relative;
        }

        .heading-container h2,
        .heading-container h3 {
            float: left;
            margin-bottom: 0;
        }

        /* Hide top links when printing */
        @media print {
            .top-link {
                display: none;
            }
        }
    </style>
</head>
<body>
    <div id="header" class="center-container">
    <h1>${headingLabel}: ${JobLabel}</h1>
    <div id="toc" class="toc">
        <div id="toctitle">${tableOfContentsLabel}</div>
        <ul class="sectlevel1">
            <li><a href="#_part_information">${partInformationLabel}</a></li>
            <li><a href="#_run_summary">${runSummaryLabel}</a></li>
            <li><a href="#_rough_stock">${roughStockLabel}</a></li>
            <li><a href="#_tool_data">${toolDataLabel}</a>
                <ul class="sectlevel2">
                    ${tool_list}
                </ul>
            </li>
            <li><a href="#_output">${outputLabel}</a></li>
            <li><a href="#_fixtures_and_workholding">${fixturesLabel}</a></li>
            <li><a href="#_squawks">${squawksLabel}</a></li>
        </ul>
    </div>


<div class="heading-container">
    <h2 id="_part_information"><a name="_part_information"></a>${partInformationLabel}</h2><a href="#header" class="top-link">Top</a>
</div>
<table>
    <colgroup>
        <col width="20%"/>
        <col width="50%"/>
        <col width="30%"/>
    </colgroup>
    <tr>
        <td>
            <strong>${PartLabel}</strong>
        </td>
        <td>
            <table>
                <colgroup>
                    <col width="50%"/>
                    <col width="50%"/>
                </colgroup>
                ${bases}
            </table>
        </td>
        <td rowspan="7" class="image-container">
            ${baseimage}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${SequenceLabel}</strong>
        </td>
        <td>
            ${Sequence}
        </td>
    </tr>

    <tr>
        <td>
            <strong>${JobTypeLabel}</strong>
        </td>
        <td>
            ${JobType}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${CADLabel}</strong>
        </td>
        <td>
            ${FileName}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${LastSaveLabel}</strong>
        </td>
        <td>
            ${LastModifiedDate}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${CustomerLabel}</strong>
        </td>
        <td>
            ${Customer}
        </td>
    </tr>
</table>

<div class="heading-container">
    <h2 id="_run_summary"><a name="_run_summary"></a>${runSummaryLabel}</h2><a href="#header" class="top-link">Top</a>
</div>
<table>
    <colgroup>
        <col width="20%"/>
        <col width="20%"/>
        <col width="20%"/>
        <col width="20%"/>
        <col width="20%"/>
    </colgroup>
    <thead>
        <tr>
            <th><strong>${opLabel}</strong></th>
            <th><strong>${jobMinZLabel}</strong></th>
            <th><strong>${jobMaxZLabel}</strong></th>
            <th><strong>${coolantLabel}</strong></th>
            <th><strong>${cycleTimeLabel}</strong></th>
        </tr>
    </thead>
    <tbody>
        ${run_summary_ops}
    </tbody>
</table>

<div class="heading-container">
    <h2 id="_rough_stock"><a name="_rough_stock"></a>${roughStockLabel}</h2><a href="#header" class="top-link">Top</a>
</div>

<table>
    <colgroup>
        <col width="30%"/>
        <col width="30%"/>
        <col width="40%"/>
    </colgroup>
    <tbody>
        <tr>
            <td><strong>${materialLabel}</strong></td>
            <td>${material}</td>
            <td rowspan="7" class="image-container">
                ${stockImage}
            </td>
        </tr>
        <tr>
            <td><strong>${sSpeedHSSLabel}</strong></td>
            <td>${surfaceSpeedHSS}</td>
        </tr>
        <tr>
            <td><strong>${sSpeedCarbideLabel}</strong></td>
        <td>${surfaceSpeedCarbide}</td>
    </tr>
    <tr>
        <td><strong>${xDimLabel}</strong></td>
        <td>${xLen}</td>
    </tr>
    <tr>
        <td><strong>${yDimLabel}</strong></td>
        <td>${yLen}</td>
    </tr>
    <tr>
        <td><strong>${zDimLabel}</strong></td>
        <td>${zLen}</td>
    </tr>
</table>


<div class="heading-container">
    <h2 id="_tool_data"><a name="_tool_data"></a>${toolDataLabel}</h2><a href="#header" class="top-link">Top</a>
</div>
${tool_data}


<div class="heading-container">
    <h2 id="_output"><a name="_output"></a>${outputLabel}</h2><a href="#header" class="top-link">Top</a>
</div>
<table cellpadding="2" cellspacing="2" bgcolor="#ffffff" style="background: #ffffff;">
    <colgroup>
        <col width="525"/>
        <col width="525"/>
    </colgroup>
    <tr>
        <td>
            <strong>${gcodeFileLabel}</strong>
        </td>
        <td>
            ${lastgcodefile}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${lastpostLabel}</strong>
        </td>
        <td>
            ${lastpostprocess}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${stopsLabel}</strong>
        </td>
        <td>
            ${optionalstops}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${programmerLabel}</strong>
        </td>
        <td>
            ${programmer}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${machineLabel}</strong>
        </td>
        <td>
            ${machine}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${postLabel}</strong>
        </td>
        <td>
            ${postprocessor}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${flagsLabel}</strong>
        </td>
        <td>
            ${postprocessorFlags}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${fileSizeLabel}</strong>
        </td>
        <td>
            ${filesize}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${lineCountLabel}</strong>
        </td>
        <td>
            ${linecount}
        </td>
    </tr>
</table>

<div class="heading-container">
    <h2 id="_fixtures_and_workholding"><a name="_fixtures_and_workholding"></a>${fixturesLabel}</h2><a href="#header" class="top-link">Top</a>
</div>
<table cellpadding="2" cellspacing="2" bgcolor="#ffffff" style="background: #ffffff;">
    <colgroup>
        <col width="525"/>
        <col width="525"/>
    </colgroup>
    <tr>
        <td>
            <strong>${offsetsLabel}</strong>
        </td>
        <td>
            ${fixtures}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${orderByLabel}</strong>
        </td>
        <td>
            ${orderBy}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${datumLabel}</strong>
        </td>
        <td class="image-container">
            ${datumImage}
        </td>
    </tr>
</table>

<div class="heading-container">
    <h2 id="_squawks"><a name="_squawks"></a>${squawksLabel}</h2><a href="#header" class="top-link">Top</a>
</div>
<table cellpadding="2" cellspacing="2" bgcolor="#ffffff" style="background-color: #ffffff;">
    <colgroup>
        <col width="100"/>
        <col width="250"/>
        <col width="250"/>
        <col width="550"/>
    </colgroup>
    <tr>
        <td>
            <strong>${noteLabel}</strong>
        </td>
        <td>
            <strong>${operatorLabel}</strong>
        </td>
        <td>
            <strong>${dateLabel}</strong>
        </td>
        <td>
            <strong>${noteLabel}</strong>
        </td>
        ${squawks}
    </tr>
</table>

<p style="line-height: 100%; margin-bottom: 0cm"><br/>
</p>
</div>
</body>
</html>
"""
)

base_template = Template(
    """
<tr>
    <td>
        ${key}
    </td>
    <td>
        ${val}
    </td>
</tr>
        """
)

squawk_template = Template(
    """
<tr>
    <td>
        ${squawkIcon}
    </td>
    <td>
        ${Operator}
    </td>
    <td>
        ${Date}
    </td>
    <td colspan="3">
        ${Note}
    </td>
</tr>
        """
)

tool_template = Template(
    """
<div class="heading-container">
    <h3 id="_tool_data_T${toolNumber}">Tool Number: T${toolNumber}</h3><a href="#header" class="top-link">Top</a>
</div>
<table>
    <colgroup>
        <col width="30%"/>
        <col width="30%"/>
        <col width="40%"/>
    </colgroup>
    <tr>
        <td>
            <strong>${descriptionLabel}</strong>
        </td>
        <td>
            ${description}
        </td>
        <td class="image-container">
            ${imagepath}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${manufLabel}</strong>
        </td>
        <td colspan="2">
            ${manufacturer}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${partNumberLabel}</strong>
        </td>
        <td colspan="2">
            ${partNumber}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${urlLabel}</strong>
        </td>
        <td colspan="2">
            ${url}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${shapeLabel}</strong>
        </td>
        <td colspan="2">
            ${shape}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${inspectionNotesLabel}</strong>
        </td>
        <td colspan="2">
            ${inspectionNotes}
        </td>
    </tr>
    <tr>
        <td>
            <strong>${diameterLabel}</strong>
        </td>
        <td colspan="2">
            ${diameter}
        </td>
    </tr>
</table>
${ops}
        """
)

op_tool_template = Template(
    """
<table cellpadding="2" cellspacing="2" bgcolor="#ffffff" style="background: #ffffff;">
    <colgroup>
        <col width="262"/>
        <col width="262"/>
        <col width="262"/>
        <col width="262"/>
    </colgroup>
    <tr>
        <td>
            <strong>${opLabel}</strong>
        </td>
        <td>
            <strong>${tcLabel}</strong>
        </td>
        <td>
            <strong>${feedLabel}</strong>
        </td>
        <td>
            <strong>${speedLabel}</strong>
        </td>
    </tr>
    <tr>
        <td>
            ${Operation}
        </td>
        <td>
            ${ToolController}
        </td>
        <td>
            ${Feed}
        </td>
        <td>
            ${Speed}
        </td>
    </tr>
</table>
        """
)

op_run_template = Template(
    """
<tr>
    <td>
        ${opName}
    </td>
    <td>
        ${minZ}
    </td>
    <td>
        ${maxZ}
    </td>
    <td>
        ${coolantMode}
    </td>
    <td>
        ${cycleTime}
    </td>
</tr>
        """
)

tool_item_template = Template(
    """
<li><a href="#_tool_data_T${toolNumber}">Tool Number: T${toolNumber}</a></li>
    """
)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM MoveView command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


class BIM_MoveView:

    def GetResources(self):
        return {
            "Pixmap": "BIM_MoveView",
            "MenuText": QT_TRANSLATE_NOOP("BIM_MoveView", "Move View"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_MoveView", "Moves this view to an existing page"),
        }

    def Activated(self):
        self.selection = FreeCADGui.Selection.getSelection()
        self.pages = FreeCAD.ActiveDocument.findObjects(Type="TechDraw::DrawPage")
        self.labels = [obj.Label for obj in self.pages]
        FreeCADGui.draftToolBar.sourceCmd = self
        FreeCADGui.draftToolBar.popupMenu(self.labels)

    def proceed(self, labelname):
        FreeCADGui.draftToolBar.sourceCmd = None
        if labelname in self.labels:
            page = self.pages[self.labels.index(labelname)]
            for view in self.selection:
                for p in view.InList:
                    if hasattr(p, "Views") and view in p.Views:
                        p.removeView(view)
                x = view.X
                y = view.Y
                page.addView(view)
                FreeCAD.ActiveDocument.recompute()
                view.X = x
                view.Y = y


FreeCADGui.addCommand("BIM_MoveView", BIM_MoveView())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import Part
import Path
import math

from FreeCAD import Vector
from CAMTests.PathTestUtils import PathTestBase


class TestPathGeom(PathTestBase):
    """Test Path <-> Wire conversion."""

    def test00(self):
        """Verify getAngle functionality."""
        self.assertRoughly(Path.Geom.getAngle(Vector(1, 0, 0)), 0)
        self.assertRoughly(Path.Geom.getAngle(Vector(1, 1, 0)), math.pi / 4)
        self.assertRoughly(Path.Geom.getAngle(Vector(0, 1, 0)), math.pi / 2)
        self.assertRoughly(Path.Geom.getAngle(Vector(-1, 1, 0)), 3 * math.pi / 4)
        self.assertRoughly(Path.Geom.getAngle(Vector(-1, 0, 0)), math.pi)
        self.assertRoughly(Path.Geom.getAngle(Vector(-1, -1, 0)), -3 * math.pi / 4)
        self.assertRoughly(Path.Geom.getAngle(Vector(0, -1, 0)), -math.pi / 2)
        self.assertRoughly(Path.Geom.getAngle(Vector(1, -1, 0)), -math.pi / 4)

    def test01(self):
        """Verify diffAngle functionality."""
        self.assertRoughly(Path.Geom.diffAngle(0, +0 * math.pi / 4, "CW") / math.pi, 0 / 4.0)
        self.assertRoughly(Path.Geom.diffAngle(0, +3 * math.pi / 4, "CW") / math.pi, 5 / 4.0)
        self.assertRoughly(Path.Geom.diffAngle(0, -3 * math.pi / 4, "CW") / math.pi, 3 / 4.0)
        self.assertRoughly(Path.Geom.diffAngle(0, +4 * math.pi / 4, "CW") / math.pi, 4 / 4.0)
        self.assertRoughly(Path.Geom.diffAngle(0, +0 * math.pi / 4, "CCW") / math.pi, 0 / 4.0)
        self.assertRoughly(Path.Geom.diffAngle(0, +3 * math.pi / 4, "CCW") / math.pi, 3 / 4.0)
        self.assertRoughly(Path.Geom.diffAngle(0, -3 * math.pi / 4, "CCW") / math.pi, 5 / 4.0)
        self.assertRoughly(Path.Geom.diffAngle(0, +4 * math.pi / 4, "CCW") / math.pi, 4 / 4.0)

        self.assertRoughly(
            Path.Geom.diffAngle(+math.pi / 4, +0 * math.pi / 4, "CW") / math.pi, 1 / 4.0
        )
        self.assertRoughly(
            Path.Geom.diffAngle(+math.pi / 4, +3 * math.pi / 4, "CW") / math.pi, 6 / 4.0
        )
        self.assertRoughly(
            Path.Geom.diffAngle(+math.pi / 4, -1 * math.pi / 4, "CW") / math.pi, 2 / 4.0
        )
        self.assertRoughly(
            Path.Geom.diffAngle(-math.pi / 4, +0 * math.pi / 4, "CW") / math.pi, 7 / 4.0
        )
        self.assertRoughly(
            Path.Geom.diffAngle(-math.pi / 4, +3 * math.pi / 4, "CW") / math.pi, 4 / 4.0
        )
        self.assertRoughly(
            Path.Geom.diffAngle(-math.pi / 4, -1 * math.pi / 4, "CW") / math.pi, 0 / 4.0
        )

        self.assertRoughly(
            Path.Geom.diffAngle(+math.pi / 4, +0 * math.pi / 4, "CCW") / math.pi,
            7 / 4.0,
        )
        self.assertRoughly(
            Path.Geom.diffAngle(+math.pi / 4, +3 * math.pi / 4, "CCW") / math.pi,
            2 / 4.0,
        )
        self.assertRoughly(
            Path.Geom.diffAngle(+math.pi / 4, -1 * math.pi / 4, "CCW") / math.pi,
            6 / 4.0,
        )
        self.assertRoughly(
            Path.Geom.diffAngle(-math.pi / 4, +0 * math.pi / 4, "CCW") / math.pi,
            1 / 4.0,
        )
        self.assertRoughly(
            Path.Geom.diffAngle(-math.pi / 4, +3 * math.pi / 4, "CCW") / math.pi,
            4 / 4.0,
        )
        self.assertRoughly(
            Path.Geom.diffAngle(-math.pi / 4, -1 * math.pi / 4, "CCW") / math.pi,
            0 / 4.0,
        )

    def test02(self):
        """Verify isVertical/isHorizontal for Vector"""
        self.assertTrue(Path.Geom.isVertical(Vector(0, 0, 1)))
        self.assertTrue(Path.Geom.isVertical(Vector(0, 0, -1)))
        self.assertFalse(Path.Geom.isVertical(Vector(1, 0, 1)))
        self.assertFalse(Path.Geom.isVertical(Vector(1, 0, -1)))

        self.assertTrue(Path.Geom.isHorizontal(Vector(1, 0, 0)))
        self.assertTrue(Path.Geom.isHorizontal(Vector(-1, 0, 0)))
        self.assertTrue(Path.Geom.isHorizontal(Vector(0, 1, 0)))
        self.assertTrue(Path.Geom.isHorizontal(Vector(0, -1, 0)))
        self.assertTrue(Path.Geom.isHorizontal(Vector(1, 1, 0)))
        self.assertTrue(Path.Geom.isHorizontal(Vector(-1, 1, 0)))
        self.assertTrue(Path.Geom.isHorizontal(Vector(1, -1, 0)))
        self.assertTrue(Path.Geom.isHorizontal(Vector(-1, -1, 0)))

        self.assertFalse(Path.Geom.isHorizontal(Vector(0, 1, 1)))
        self.assertFalse(Path.Geom.isHorizontal(Vector(0, -1, 1)))
        self.assertFalse(Path.Geom.isHorizontal(Vector(0, 1, -1)))
        self.assertFalse(Path.Geom.isHorizontal(Vector(0, -1, -1)))

    def test03(self):
        """Verify isVertical/isHorizontal for Edges"""

        # lines
        self.assertTrue(
            Path.Geom.isVertical(Part.Edge(Part.LineSegment(Vector(-1, -1, -1), Vector(-1, -1, 8))))
        )
        self.assertFalse(
            Path.Geom.isVertical(Part.Edge(Part.LineSegment(Vector(-1, -1, -1), Vector(1, -1, 8))))
        )
        self.assertFalse(
            Path.Geom.isVertical(Part.Edge(Part.LineSegment(Vector(-1, -1, -1), Vector(-1, 1, 8))))
        )

        self.assertTrue(
            Path.Geom.isHorizontal(
                Part.Edge(Part.LineSegment(Vector(1, -1, -1), Vector(-1, -1, -1)))
            )
        )
        self.assertTrue(
            Path.Geom.isHorizontal(
                Part.Edge(Part.LineSegment(Vector(-1, 1, -1), Vector(-1, -1, -1)))
            )
        )
        self.assertTrue(
            Path.Geom.isHorizontal(
                Part.Edge(Part.LineSegment(Vector(1, 1, -1), Vector(-1, -1, -1)))
            )
        )
        self.assertFalse(
            Path.Geom.isHorizontal(Part.Edge(Part.LineSegment(Vector(1, -1, -1), Vector(1, -1, 8))))
        )
        self.assertFalse(
            Path.Geom.isHorizontal(Part.Edge(Part.LineSegment(Vector(-1, 1, -1), Vector(-1, 1, 8))))
        )

        # circles
        self.assertTrue(
            Path.Geom.isVertical(Part.Edge(Part.makeCircle(4, Vector(), Vector(0, 1, 0))))
        )
        self.assertTrue(
            Path.Geom.isVertical(Part.Edge(Part.makeCircle(4, Vector(), Vector(1, 0, 0))))
        )
        self.assertTrue(
            Path.Geom.isVertical(Part.Edge(Part.makeCircle(4, Vector(), Vector(1, 1, 0))))
        )
        self.assertFalse(
            Path.Geom.isVertical(Part.Edge(Part.makeCircle(4, Vector(), Vector(1, 1, 1))))
        )

        self.assertTrue(
            Path.Geom.isHorizontal(Part.Edge(Part.makeCircle(4, Vector(), Vector(0, 0, 1))))
        )
        self.assertFalse(
            Path.Geom.isHorizontal(Part.Edge(Part.makeCircle(4, Vector(), Vector(0, 1, 1))))
        )
        self.assertFalse(
            Path.Geom.isHorizontal(Part.Edge(Part.makeCircle(4, Vector(), Vector(1, 0, 1))))
        )
        self.assertFalse(
            Path.Geom.isHorizontal(Part.Edge(Part.makeCircle(4, Vector(), Vector(1, 1, 1))))
        )

        # bezier curves
        # ml: I know nothing about bezier curves, so this might be bollocks
        # and now I disable the tests because they seem to fail on OCE
        # bezier = Part.BezierCurve()
        # bezier.setPoles([Vector(), Vector(1,1,0), Vector(2,1,0), Vector(2,2,0)])
        # self.assertTrue(Path.Geom.isHorizontal(Part.Edge(bezier)))
        # self.assertFalse(Path.Geom.isVertical(Part.Edge(bezier)))
        # bezier.setPoles([Vector(), Vector(1,1,1), Vector(2,1,0), Vector(2,2,0)])
        # self.assertFalse(Path.Geom.isHorizontal(Part.Edge(bezier)))
        # self.assertFalse(Path.Geom.isVertical(Part.Edge(bezier)))
        # bezier.setPoles([Vector(), Vector(1,1,0), Vector(2,1,1), Vector(2,2,0)])
        # self.assertFalse(Path.Geom.isHorizontal(Part.Edge(bezier)))
        # self.assertFalse(Path.Geom.isVertical(Part.Edge(bezier)))
        # bezier.setPoles([Vector(), Vector(1,1,0), Vector(2,1,0), Vector(2,2,1)])
        # self.assertFalse(Path.Geom.isHorizontal(Part.Edge(bezier)))
        # self.assertFalse(Path.Geom.isVertical(Part.Edge(bezier)))
        #
        # bezier.setPoles([Vector(), Vector(1,1,1), Vector(2,2,2), Vector(0,0,3)])
        # self.assertFalse(Path.Geom.isHorizontal(Part.Edge(bezier)))
        # self.assertTrue(Path.Geom.isVertical(Part.Edge(bezier)))

    def test04(self):
        """Verify isVertical/isHorizontal for faces"""

        # planes
        xPlane = Part.makePlane(100, 100, Vector(), Vector(1, 0, 0))
        yPlane = Part.makePlane(100, 100, Vector(), Vector(0, 1, 0))
        zPlane = Part.makePlane(100, 100, Vector(), Vector(0, 0, 1))
        xyPlane = Part.makePlane(100, 100, Vector(), Vector(1, 1, 0))
        xzPlane = Part.makePlane(100, 100, Vector(), Vector(1, 0, 1))
        yzPlane = Part.makePlane(100, 100, Vector(), Vector(0, 1, 1))

        self.assertTrue(Path.Geom.isVertical(xPlane))
        self.assertTrue(Path.Geom.isVertical(yPlane))
        self.assertFalse(Path.Geom.isVertical(zPlane))
        self.assertTrue(Path.Geom.isVertical(xyPlane))
        self.assertFalse(Path.Geom.isVertical(xzPlane))
        self.assertFalse(Path.Geom.isVertical(yzPlane))

        self.assertFalse(Path.Geom.isHorizontal(xPlane))
        self.assertFalse(Path.Geom.isHorizontal(yPlane))
        self.assertTrue(Path.Geom.isHorizontal(zPlane))
        self.assertFalse(Path.Geom.isHorizontal(xyPlane))
        self.assertFalse(Path.Geom.isHorizontal(xzPlane))
        self.assertFalse(Path.Geom.isHorizontal(yzPlane))

        # cylinders
        xCylinder = [
            f
            for f in Part.makeCylinder(1, 1, Vector(), Vector(1, 0, 0)).Faces
            if type(f.Surface) == Part.Cylinder
        ][0]
        yCylinder = [
            f
            for f in Part.makeCylinder(1, 1, Vector(), Vector(0, 1, 0)).Faces
            if type(f.Surface) == Part.Cylinder
        ][0]
        zCylinder = [
            f
            for f in Part.makeCylinder(1, 1, Vector(), Vector(0, 0, 1)).Faces
            if type(f.Surface) == Part.Cylinder
        ][0]
        xyCylinder = [
            f
            for f in Part.makeCylinder(1, 1, Vector(), Vector(1, 1, 0)).Faces
            if type(f.Surface) == Part.Cylinder
        ][0]
        xzCylinder = [
            f
            for f in Part.makeCylinder(1, 1, Vector(), Vector(1, 0, 1)).Faces
            if type(f.Surface) == Part.Cylinder
        ][0]
        yzCylinder = [
            f
            for f in Part.makeCylinder(1, 1, Vector(), Vector(0, 1, 1)).Faces
            if type(f.Surface) == Part.Cylinder
        ][0]

        self.assertTrue(Path.Geom.isHorizontal(xCylinder))
        self.assertTrue(Path.Geom.isHorizontal(yCylinder))
        self.assertFalse(Path.Geom.isHorizontal(zCylinder))
        self.assertTrue(Path.Geom.isHorizontal(xyCylinder))
        self.assertFalse(Path.Geom.isHorizontal(xzCylinder))
        self.assertFalse(Path.Geom.isHorizontal(yzCylinder))

    def test07(self):
        """Verify speed interpolation works for different pitches"""
        # horizontal
        self.assertRoughly(100, Path.Geom.speedBetweenPoints(Vector(), Vector(1, 1, 0), 100, 50))
        self.assertRoughly(100, Path.Geom.speedBetweenPoints(Vector(1, 1, 0), Vector(), 100, 50))
        # vertical
        self.assertRoughly(50, Path.Geom.speedBetweenPoints(Vector(), Vector(0, 0, 1), 100, 50))
        self.assertRoughly(50, Path.Geom.speedBetweenPoints(Vector(0, 0, 1), Vector(), 100, 50))
        # 45°
        self.assertRoughly(75, Path.Geom.speedBetweenPoints(Vector(), Vector(1, 0, 1), 100, 50))
        self.assertRoughly(75, Path.Geom.speedBetweenPoints(Vector(), Vector(0, 1, 1), 100, 50))
        self.assertRoughly(
            75,
            Path.Geom.speedBetweenPoints(Vector(), Vector(0.707, 0.707, 1), 100, 50),
            0.01,
        )
        self.assertRoughly(75, Path.Geom.speedBetweenPoints(Vector(1, 0, 1), Vector(), 100, 50))
        self.assertRoughly(75, Path.Geom.speedBetweenPoints(Vector(0, 1, 1), Vector(), 100, 50))
        self.assertRoughly(
            75,
            Path.Geom.speedBetweenPoints(Vector(0.707, 0.707, 1), Vector(), 100, 50),
            0.01,
        )
        # 30°
        self.assertRoughly(
            66.66,
            Path.Geom.speedBetweenPoints(Vector(), Vector(0.5774, 0, 1), 100, 50),
            0.01,
        )
        self.assertRoughly(
            66.66,
            Path.Geom.speedBetweenPoints(Vector(), Vector(0, 0.5774, 1), 100, 50),
            0.01,
        )
        self.assertRoughly(
            66.66,
            Path.Geom.speedBetweenPoints(Vector(0.5774, 0, 1), Vector(), 100, 50),
            0.01,
        )
        self.assertRoughly(
            66.66,
            Path.Geom.speedBetweenPoints(Vector(0, 0.5774, 1), Vector(), 100, 50),
            0.01,
        )
        # 60°
        self.assertRoughly(
            83.33,
            Path.Geom.speedBetweenPoints(Vector(), Vector(1, 0, 0.5774), 100, 50),
            0.01,
        )
        self.assertRoughly(
            83.33,
            Path.Geom.speedBetweenPoints(Vector(), Vector(0, 1, 0.5774), 100, 50),
            0.01,
        )
        self.assertRoughly(
            83.33,
            Path.Geom.speedBetweenPoints(Vector(1, 0, 0.5774), Vector(), 100, 50),
            0.01,
        )
        self.assertRoughly(
            83.33,
            Path.Geom.speedBetweenPoints(Vector(0, 1, 0.5774), Vector(), 100, 50),
            0.01,
        )

    def test08(self):
        """Verify speed interpolation works for different pitches if vSpeed > hSpeed"""
        # horizontal
        self.assertRoughly(50, Path.Geom.speedBetweenPoints(Vector(), Vector(1, 1, 0), 50, 100))
        self.assertRoughly(50, Path.Geom.speedBetweenPoints(Vector(1, 1, 0), Vector(), 50, 100))
        # vertical
        self.assertRoughly(100, Path.Geom.speedBetweenPoints(Vector(), Vector(0, 0, 1), 50, 100))
        self.assertRoughly(100, Path.Geom.speedBetweenPoints(Vector(0, 0, 1), Vector(), 50, 100))
        # 45°
        self.assertRoughly(75, Path.Geom.speedBetweenPoints(Vector(), Vector(1, 0, 1), 50, 100))
        self.assertRoughly(75, Path.Geom.speedBetweenPoints(Vector(), Vector(0, 1, 1), 50, 100))
        self.assertRoughly(
            75,
            Path.Geom.speedBetweenPoints(Vector(), Vector(0.707, 0.707, 1), 50, 100),
            0.01,
        )
        self.assertRoughly(75, Path.Geom.speedBetweenPoints(Vector(1, 0, 1), Vector(), 50, 100))
        self.assertRoughly(75, Path.Geom.speedBetweenPoints(Vector(0, 1, 1), Vector(), 50, 100))
        self.assertRoughly(
            75,
            Path.Geom.speedBetweenPoints(Vector(0.707, 0.707, 1), Vector(), 50, 100),
            0.01,
        )
        # 30°
        self.assertRoughly(
            83.33,
            Path.Geom.speedBetweenPoints(Vector(), Vector(0.5774, 0, 1), 50, 100),
            0.01,
        )
        self.assertRoughly(
            83.33,
            Path.Geom.speedBetweenPoints(Vector(), Vector(0, 0.5774, 1), 50, 100),
            0.01,
        )
        self.assertRoughly(
            83.33,
            Path.Geom.speedBetweenPoints(Vector(0.5774, 0, 1), Vector(), 50, 100),
            0.01,
        )
        self.assertRoughly(
            83.33,
            Path.Geom.speedBetweenPoints(Vector(0, 0.5774, 1), Vector(), 50, 100),
            0.01,
        )
        # 60°
        self.assertRoughly(
            66.66,
            Path.Geom.speedBetweenPoints(Vector(), Vector(1, 0, 0.5774), 50, 100),
            0.01,
        )
        self.assertRoughly(
            66.66,
            Path.Geom.speedBetweenPoints(Vector(), Vector(0, 1, 0.5774), 50, 100),
            0.01,
        )
        self.assertRoughly(
            66.66,
            Path.Geom.speedBetweenPoints(Vector(1, 0, 0.5774), Vector(), 50, 100),
            0.01,
        )
        self.assertRoughly(
            66.66,
            Path.Geom.speedBetweenPoints(Vector(0, 1, 0.5774), Vector(), 50, 100),
            0.01,
        )

    def test10(self):
        """Verify proper geometry objects for G1 and G01 commands are created."""
        spt = Vector(1, 2, 3)
        self.assertLine(
            Path.Geom.edgeForCmd(Path.Command("G1", {"X": 7, "Y": 2, "Z": 3}), spt),
            spt,
            Vector(7, 2, 3),
        )
        self.assertLine(
            Path.Geom.edgeForCmd(Path.Command("G01", {"X": 1, "Y": 3, "Z": 5}), spt),
            spt,
            Vector(1, 3, 5),
        )

    def test20(self):
        """Verify proper geometry for arcs in the XY-plane are created."""
        p1 = Vector(0, -1, 2)
        p2 = Vector(-1, 0, 2)
        self.assertArc(
            Path.Geom.edgeForCmd(
                Path.Command("G2", {"X": p2.x, "Y": p2.y, "Z": p2.z, "I": 0, "J": 1, "K": 0}),
                p1,
            ),
            p1,
            p2,
            "CW",
        )
        self.assertArc(
            Path.Geom.edgeForCmd(
                Path.Command("G3", {"X": p1.x, "Y": p1.y, "z": p1.z, "I": -1, "J": 0, "K": 0}),
                p2,
            ),
            p2,
            p1,
            "CCW",
        )

    def test30(self):
        """Verify proper geometry for arcs with rising and fall ing Z-axis are created."""
        # print("------ rising helix -------")
        p1 = Vector(0, 1, 0)
        p2 = Vector(1, 0, 2)
        self.assertCurve(
            Path.Geom.edgeForCmd(
                Path.Command("G2", {"X": p2.x, "Y": p2.y, "Z": p2.z, "I": 0, "J": -1, "K": 1}),
                p1,
            ),
            p1,
            Vector(1 / math.sqrt(2), 1 / math.sqrt(2), 1),
            p2,
        )
        p1 = Vector(-1, 0, 0)
        p2 = Vector(0, -1, 2)
        self.assertCurve(
            Path.Geom.edgeForCmd(
                Path.Command("G3", {"X": p2.x, "Y": p2.y, "Z": p2.z, "I": 1, "J": 0, "K": 1}),
                p1,
            ),
            p1,
            Vector(-1 / math.sqrt(2), -1 / math.sqrt(2), 1),
            p2,
        )

        # print("------ falling helix -------")
        p1 = Vector(0, -1, 2)
        p2 = Vector(-1, 0, 0)
        self.assertCurve(
            Path.Geom.edgeForCmd(
                Path.Command("G2", {"X": p2.x, "Y": p2.y, "Z": p2.z, "I": 0, "J": 1, "K": -1}),
                p1,
            ),
            p1,
            Vector(-1 / math.sqrt(2), -1 / math.sqrt(2), 1),
            p2,
        )
        p1 = Vector(-1, 0, 2)
        p2 = Vector(0, -1, 0)
        self.assertCurve(
            Path.Geom.edgeForCmd(
                Path.Command("G3", {"X": p2.x, "Y": p2.y, "Z": p2.z, "I": 1, "J": 0, "K": -1}),
                p1,
            ),
            p1,
            Vector(-1 / math.sqrt(2), -1 / math.sqrt(2), 1),
            p2,
        )

    def test40(self):
        """Verify arc results in proper G2/3 command."""
        p1 = Vector(0, -10, 0)
        p2 = Vector(-10, 0, 0)
        p3 = Vector(0, +10, 0)
        p4 = Vector(+10, 0, 0)

        def cmds(pa, pb, pc, flip):
            return Path.Geom.cmdsForEdge(Part.Edge(Part.Arc(pa, pb, pc)), flip)[0]

        def cmd(g, end, off):
            return Path.Command(
                g,
                {
                    "X": end.x,
                    "Y": end.y,
                    "Z": end.z,
                    "I": off.x,
                    "J": off.y,
                    "K": off.z,
                },
            )

        self.assertCommandEqual(cmds(p1, p2, p3, False), cmd("G2", p3, Vector(0, 10, 0)))
        self.assertCommandEqual(cmds(p1, p4, p3, False), cmd("G3", p3, Vector(0, 10, 0)))

        self.assertCommandEqual(cmds(p1, p2, p3, True), cmd("G3", p1, Vector(0, -10, 0)))
        self.assertCommandEqual(cmds(p1, p4, p3, True), cmd("G2", p1, Vector(0, -10, 0)))

    def test41(self):
        """Verify circle results in proper G2/G3 commands."""

        def cmds(center, radius, up=True):
            norm = Vector(0, 0, 1) if up else Vector(0, 0, -1)
            return Path.Geom.cmdsForEdge(Part.Edge(Part.Circle(center, norm, radius)))[0]

        def cmd(g, end, off):
            return Path.Command(
                g,
                {
                    "X": end.x,
                    "Y": end.y,
                    "Z": end.z,
                    "I": off.x,
                    "J": off.y,
                    "K": off.z,
                },
            )

        center = Vector(10, 10, 0)
        radius = 5

        self.assertCommandEqual(
            cmds(center, radius), cmd("G3", Vector(15, 10, 0), Vector(-5, 0, 0))
        )

    def test42(self):
        """Verify ellipsis results in a proper segmentation of G1 commands."""
        ellipse = Part.Edge(Part.Ellipse())
        cmds = Path.Geom.cmdsForEdge(ellipse)
        # let's make sure all commands are G1 and there are more than 20 of those
        self.assertGreater(len(cmds), 20)
        self.assertTrue(all([cmd.Name == "G1" for cmd in cmds]))

    def test50(self):
        """Verify proper wire(s) aggregation from a Path."""
        commands = []
        commands.append(Path.Command("G1", {"X": 1}))
        commands.append(Path.Command("G1", {"Y": 1}))
        commands.append(Path.Command("G0", {"X": 0}))
        commands.append(Path.Command("G1", {"Y": 0}))

        wire, rapid, rapid_indexes = Path.Geom.wireForPath(Path.Path(commands))
        self.assertEqual(len(wire.Edges), 4)
        self.assertLine(wire.Edges[0], Vector(0, 0, 0), Vector(1, 0, 0))
        self.assertLine(wire.Edges[1], Vector(1, 0, 0), Vector(1, 1, 0))
        self.assertLine(wire.Edges[2], Vector(1, 1, 0), Vector(0, 1, 0))
        self.assertLine(wire.Edges[3], Vector(0, 1, 0), Vector(0, 0, 0))
        self.assertEqual(len(rapid), 1)
        self.assertTrue(Path.Geom.edgesMatch(rapid[0], wire.Edges[2]))

        wires = Path.Geom.wiresForPath(Path.Path(commands))
        self.assertEqual(len(wires), 2)
        self.assertEqual(len(wires[0].Edges), 2)
        self.assertLine(wires[0].Edges[0], Vector(0, 0, 0), Vector(1, 0, 0))
        self.assertLine(wires[0].Edges[1], Vector(1, 0, 0), Vector(1, 1, 0))
        self.assertEqual(len(wires[1].Edges), 1)
        self.assertLine(wires[1].Edges[0], Vector(0, 1, 0), Vector(0, 0, 0))

    def test60(self):
        """Verify arcToHelix returns proper helix."""
        p1 = Vector(10, -10, 0)
        p2 = Vector(0, 0, 0)
        p3 = Vector(10, 10, 0)

        e = Path.Geom.arcToHelix(Part.Edge(Part.Arc(p1, p2, p3)), 0, 2)
        self.assertCurve(e, p1, p2 + Vector(0, 0, 1), p3 + Vector(0, 0, 2))

        e = Path.Geom.arcToHelix(Part.Edge(Part.Arc(p1, p2, p3)), 3, 7)
        self.assertCurve(e, p1 + Vector(0, 0, 3), p2 + Vector(0, 0, 5), p3 + Vector(0, 0, 7))

        e = Path.Geom.arcToHelix(Part.Edge(Part.Arc(p1, p2, p3)), 9, 1)
        self.assertCurve(e, p1 + Vector(0, 0, 9), p2 + Vector(0, 0, 5), p3 + Vector(0, 0, 1))

        dz = Vector(0, 0, 3)
        p11 = p1 + dz
        p12 = p2 + dz
        p13 = p3 + dz

        e = Path.Geom.arcToHelix(Part.Edge(Part.Arc(p11, p12, p13)), 0, 8)
        self.assertCurve(e, p1, p2 + Vector(0, 0, 4), p3 + Vector(0, 0, 8))

        e = Path.Geom.arcToHelix(Part.Edge(Part.Arc(p11, p12, p13)), 2, -2)
        self.assertCurve(e, p1 + Vector(0, 0, 2), p2, p3 + Vector(0, 0, -2))

        p1 = Vector(10, -10, 1)
        p2 = Vector(10 - 10 * math.sin(math.pi / 4), -10 * math.cos(math.pi / 4), 1)
        p3 = Vector(0, 0, 1)
        e = Path.Geom.arcToHelix(Part.Edge(Part.Arc(p1, p2, p3)), 0, 5)
        self.assertCurve(e, Vector(10, -10, 0), Vector(p2.x, p2.y, 2.5), Vector(0, 0, 5))

    def test62(self):
        """Verify splitArcAt returns proper subarcs."""
        p1 = Vector(10, -10, 0)
        p2 = Vector(0, 0, 0)
        p3 = Vector(10, 10, 0)

        arc = Part.Edge(Part.Arc(p1, p2, p3))

        o = 10 * math.sin(math.pi / 4)
        p12 = Vector(10 - o, -o, 0)
        p23 = Vector(10 - o, +o, 0)

        e = Path.Geom.splitArcAt(arc, p2)
        self.assertCurve(e[0], p1, p12, p2)
        self.assertCurve(e[1], p2, p23, p3)

        p34 = Vector(10 - 10 * math.sin(1 * math.pi / 8), -10 * math.cos(1 * math.pi / 8), 0)
        p45 = Vector(10 - 10 * math.sin(5 * math.pi / 8), -10 * math.cos(5 * math.pi / 8), 0)

        e = Path.Geom.splitArcAt(arc, p12)
        self.assertCurve(e[0], p1, p34, p12)
        self.assertCurve(e[1], p12, p45, p3)

    def test65(self):
        """Verify splitEdgeAt."""

        # split a line segment
        e = Path.Geom.splitEdgeAt(
            Part.Edge(Part.LineSegment(Vector(), Vector(2, 4, 6))), Vector(1, 2, 3)
        )
        self.assertLine(e[0], Vector(), Vector(1, 2, 3))
        self.assertLine(e[1], Vector(1, 2, 3), Vector(2, 4, 6))

        # split an arc
        p1 = Vector(10, -10, 1)
        p2 = Vector(0, 0, 1)
        p3 = Vector(10, 10, 1)
        arc = Part.Edge(Part.Arc(p1, p2, p3))
        e = Path.Geom.splitEdgeAt(arc, p2)
        o = 10 * math.sin(math.pi / 4)
        p12 = Vector(10 - o, -o, 1)
        p23 = Vector(10 - o, +o, 1)
        self.assertCurve(e[0], p1, p12, p2)
        self.assertCurve(e[1], p2, p23, p3)

        # split a helix
        p1 = Vector(10, -10, 0)
        p2 = Vector(0, 0, 5)
        p3 = Vector(10, 10, 10)
        h = Path.Geom.arcToHelix(arc, 0, 10)
        self.assertCurve(h, p1, p2, p3)

        e = Path.Geom.splitEdgeAt(h, p2)
        o = 10 * math.sin(math.pi / 4)
        p12 = Vector(10 - o, -o, 2.5)
        p23 = Vector(10 - o, +o, 7.5)
        self.assertCurve(e[0], p1, p12, p2)
        self.assertCurve(e[1], p2, p23, p3)

    def test66(self):
        """Split arc real world sample"""

        af = Vector(421.55, 378.41, 1)
        am = Vector(459.51, 372.61, 1)
        al = Vector(491.75, 351.75, 1)
        arc = Part.Edge(Part.ArcOfCircle(af, am, al))
        ac = arc.Curve.Center

        s = Vector(434.54, 378.26, 1)
        head, tail = Path.Geom.splitEdgeAt(arc, s)

        # make sure the arcs connect as they should
        self.assertCoincide(
            arc.valueAt(arc.FirstParameter), head.valueAt(head.FirstParameter), 0.005
        )
        self.assertCoincide(s, head.valueAt(head.LastParameter), 0.005)
        self.assertCoincide(s, tail.valueAt(tail.FirstParameter), 0.005)
        i = arc.valueAt(arc.LastParameter)
        j = tail.valueAt(tail.LastParameter)
        print("(%.2f, %.2f, %.2f) vs. (%.2f, %.2f, %.2f)" % (i.x, i.y, i.z, j.x, j.y, j.z))
        self.assertCoincide(arc.valueAt(arc.LastParameter), tail.valueAt(tail.LastParameter), 0.005)

        # make sure the radii match
        self.assertRoughly(arc.Curve.Radius, head.Curve.Radius, 0.001)
        self.assertRoughly(arc.Curve.Radius, tail.Curve.Radius, 0.001)

        # also, all arcs should have the same center
        self.assertCoincide(arc.Curve.Center, head.Curve.Center, 0.001)
        self.assertCoincide(arc.Curve.Center, tail.Curve.Center, 0.001)

    def test70(self):
        """Flip a line."""
        edge = Part.Edge(Part.Line(Vector(0, 0, 0), Vector(3, 2, 1)))
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))
        edge = Part.Edge(Part.Line(Vector(0, 0, 0), Vector(-3, -2, -1)))
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

    def test71(self):
        """Flip a line segment."""
        edge = Part.Edge(Part.LineSegment(Vector(0, 0, 0), Vector(3, 2, 1)))
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))
        edge = Part.Edge(Part.LineSegment(Vector(4, 2, 1), Vector(-3, -7, 9)))
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

        edge = Part.makeLine(Vector(1, 0, 3), Vector(3, 2, 1))
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

    def test72(self):
        """Flip a circle"""
        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, 1))
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, -1))
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

    def test73(self):
        """Flip an arc"""
        # make sure all 4 quadrants work
        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, 1), 45, 90)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, 1), 100, 170)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, 1), 200, 250)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, 1), 300, 340)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

        # and the other way around too
        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, -1), 45, 90)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))
        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, -1), 100, 170)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))
        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, -1), 200, 250)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))
        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, -1), 300, 340)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

    def test74(self):
        """Flip a rotated arc"""
        # oh yes ...
        edge = Part.makeCircle(3, Vector(1, 3, 2), Vector(0, 0, 1), 45, 90)
        edge.rotate(edge.Curve.Center, Vector(0, 0, 1), -90)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

    def test75(self):
        """Flip a B-spline"""
        spline = Part.BSplineCurve()
        spline.interpolate([Vector(1, 2, 3), Vector(-3, 0, 7), Vector(-3, 1, 9), Vector(1, 3, 5)])
        edge = Part.Edge(spline)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

        edge = Part.Edge(spline, 4, 12)
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

        edge = Part.Edge(
            Part.BSplineCurve(
                [
                    Vector(-8, 4, 0),
                    Vector(1, -5, 0),
                    Vector(5, 11, 0),
                    Vector(12, -5, 0),
                ],
                weights=[2, 3, 5, 7],
            )
        )
        self.assertEdgeShapesMatch(edge, Path.Geom.flipEdge(edge))

    def test76(self):
        """Flip an offset wire"""

        e0 = Part.Edge(
            Part.BSplineCurve(
                [
                    Vector(-8, 4, 0),
                    Vector(1, -5, 0),
                    Vector(5, 11, 0),
                    Vector(12, -5, 0),
                ],
                weights=[2, 3, 5, 7],
            )
        )
        e1 = Part.Edge(Part.LineSegment(Vector(12, -5, 0), Vector(0, -7, 0)))
        e2 = Part.Edge(Part.LineSegment(Vector(0, -7, 0), Vector(-8, 4, 0)))
        w0 = Part.Wire([e0, e1, e2])
        w1 = w0.makeOffset2D(1)
        w2 = Path.Geom.flipWire(w1)
        # do some sanity checks
        self.assertTrue(w2.isValid())
        self.assertTrue(w2.isClosed())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import json
import Path
from typing import Mapping, List, Optional, cast
import FreeCAD
from ...assets import Asset, AssetUri, AssetSerializer
from ...shape import ToolBitShape
from ..models.base import ToolBit


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class FCTBSerializer(AssetSerializer):
    for_class = ToolBit
    mime_type = "application/x-freecad-toolbit"
    extensions = (".fctb",)

    @classmethod
    def get_label(cls) -> str:
        return FreeCAD.Qt.translate("CAM", "FreeCAD Tool")

    @classmethod
    def extract_dependencies(cls, data: bytes) -> List[AssetUri]:
        """Extracts URIs of dependencies from serialized data."""
        Path.Log.debug(f"FCTBSerializer.extract_dependencies: raw data = {data!r}")
        data_dict = json.loads(data.decode("utf-8"))
        shape = data_dict["shape"]
        return [ToolBitShape.resolve_name(shape)]

    @classmethod
    def serialize(cls, asset: Asset) -> bytes:
        # Ensure the asset is a ToolBit instance before serializing
        if not isinstance(asset, ToolBit):
            raise TypeError(f"Expected ToolBit instance, got {type(asset).__name__}")
        attrs = asset.to_dict()
        return json.dumps(attrs, sort_keys=True, indent=2).encode("utf-8")

    @classmethod
    def deserialize(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
    ) -> ToolBit:
        """
        Creates a ToolBit instance from serialized data and resolved
        dependencies.
        """
        attrs = json.loads(data.decode("utf-8", "ignore"))
        attrs["id"] = id  # Ensure id is available for from_dict

        if dependencies is None:
            # Shallow load: dependencies are not resolved.
            # Delegate to from_dict with shallow=True.
            Path.Log.debug(f"FCTBSerializer.deserialize: shallow. id = {id!r}, attrs = {attrs!r}")
            return ToolBit.from_dict(attrs, shallow=True)

        # Full load: dependencies are resolved.
        # Proceed with existing logic to use the resolved shape.
        shape_id = attrs.get("shape")
        if not shape_id:
            Path.Log.warning("ToolBit data is missing 'shape' key, defaulting to 'endmill'")
            shape_id = "endmill"

        shape_uri = ToolBitShape.resolve_name(shape_id)
        shape = dependencies.get(shape_uri)

        if shape is None:
            raise ValueError(
                f"Dependency for shape '{shape_id}' not found by uri {shape_uri}" f" {dependencies}"
            )
        elif not isinstance(shape, ToolBitShape):
            raise ValueError(
                f"Dependency for shape '{shape_id}' found by uri {shape_uri} "
                f"is not a ToolBitShape instance. {dependencies}"
            )

        return ToolBit.from_shape(shape, attrs, id)

    @classmethod
    def deep_deserialize(cls, data: bytes) -> ToolBit:
        """Deep deserialize preserving the original toolbit ID."""

        attrs_map = json.loads(data)
        original_id = attrs_map.get("id")

        asset_class = cast(ToolBit, cls.for_class)
        toolbit = asset_class.from_dict(attrs_map)

        if original_id:
            toolbit.id = original_id  # Preserve the original ID

        return toolbit



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.SetupSheetOpPrototype as PathSetupSheetOpPrototype

from PySide import QtCore, QtGui

__title__ = "CAM Property Editor"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Task panel editor for Properties"


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class _PropertyEditor(object):
    """Base class of all property editors - just outlines the TableView delegate interface."""

    def __init__(self, obj, prop):
        self.obj = obj
        self.prop = prop

    def widget(self, parent):
        """widget(parent) ... called by the delegate to get a new editor widget.
        Must be implemented by subclasses and return the widget."""
        pass

    def setEditorData(self, widget):
        """setEditorData(widget) ... called by the delegate to initialize the editor.
        The widget is the object returned by widget().
        Must be implemented by subclasses."""
        pass

    def setModelData(self, widget):
        """setModelData(widget) ... called by the delegate to store new values.
        Must be implemented by subclasses."""
        pass

    def propertyValue(self):
        return self.obj.getPropertyByName(self.prop)

    def setProperty(self, value):
        setattr(self.obj, self.prop, value)

    def displayString(self):
        return self.propertyValue()


class _PropertyEditorBool(_PropertyEditor):
    """Editor for boolean values - uses a combo box."""

    def widget(self, parent):
        return QtGui.QComboBox(parent)

    def setEditorData(self, widget):
        widget.clear()
        widget.addItems([str(False), str(True)])
        index = 1 if self.propertyValue() else 0
        widget.setCurrentIndex(index)

    def setModelData(self, widget):
        self.setProperty(widget.currentText() == str(True))


class _PropertyEditorString(_PropertyEditor):
    """Editor for string values - uses a line edit."""

    def widget(self, parent):
        return QtGui.QLineEdit(parent)

    def setEditorData(self, widget):
        text = "" if self.propertyValue() is None else self.propertyValue()
        widget.setText(text)

    def setModelData(self, widget):
        self.setProperty(widget.text())


class _PropertyEditorQuantity(_PropertyEditor):
    def widget(self, parent):
        return QtGui.QLineEdit(parent)

    def setEditorData(self, widget):
        quantity = self.propertyValue()
        if quantity is None:
            quantity = self.defaultQuantity()
        widget.setText(quantity.getUserPreferred()[0])

    def defaultQuantity(self):
        pass

    def setModelData(self, widget):
        self.setProperty(FreeCAD.Units.Quantity(widget.text()))

    def displayString(self):
        if self.propertyValue() is None:
            return ""
        return self.propertyValue().getUserPreferred()[0]


class _PropertyEditorAngle(_PropertyEditorQuantity):
    """Editor for angle values - uses a line edit"""

    def defaultQuantity(self):
        return FreeCAD.Units.Quantity(0, FreeCAD.Units.Angle)


class _PropertyEditorLength(_PropertyEditorQuantity):
    """Editor for length values - uses a line edit."""

    def defaultQuantity(self):
        return FreeCAD.Units.Quantity(0, FreeCAD.Units.Length)


class _PropertyEditorPercent(_PropertyEditor):
    """Editor for percent values - uses a spin box."""

    def widget(self, parent):
        return QtGui.QSpinBox(parent)

    def setEditorData(self, widget):
        widget.setRange(0, 100)
        value = self.propertyValue()
        if value is None:
            value = 0
        widget.setValue(value)

    def setModelData(self, widget):
        self.setProperty(widget.value())


class _PropertyEditorInteger(_PropertyEditor):
    """Editor for integer values - uses a spin box."""

    def widget(self, parent):
        return QtGui.QSpinBox(parent)

    def setEditorData(self, widget):
        value = self.propertyValue()
        if value is None:
            value = 0
        widget.setValue(value)

    def setModelData(self, widget):
        self.setProperty(widget.value())


class _PropertyEditorFloat(_PropertyEditor):
    """Editor for float values - uses a double spin box."""

    def widget(self, parent):
        return QtGui.QDoubleSpinBox(parent)

    def setEditorData(self, widget):
        value = self.propertyValue()
        if value is None:
            value = 0.0
        widget.setValue(value)

    def setModelData(self, widget):
        self.setProperty(widget.value())


class _PropertyEditorFile(_PropertyEditor):
    def widget(self, parent):
        return QtGui.QLineEdit(parent)

    def setEditorData(self, widget):
        text = "" if self.propertyValue() is None else self.propertyValue()
        widget.setText(text)

    def setModelData(self, widget):
        self.setProperty(widget.text())


class _PropertyEditorEnumeration(_PropertyEditor):
    def widget(self, parent):
        return QtGui.QComboBox(parent)

    def setEditorData(self, widget):
        widget.clear()
        widget.addItems(self.obj.getEnumerationsOfProperty(self.prop))
        widget.setCurrentText(self.propertyValue())

    def setModelData(self, widget):
        self.setProperty(widget.currentText())


_EditorFactory = {
    "App::PropertyAngle": _PropertyEditorAngle,
    "App::PropertyBool": _PropertyEditorBool,
    "App::PropertyDistance": _PropertyEditorLength,
    "App::PropertyEnumeration": _PropertyEditorEnumeration,
    #'App::PropertyFile'        : _PropertyEditorFile,
    "App::PropertyFloat": _PropertyEditorFloat,
    "App::PropertyInteger": _PropertyEditorInteger,
    "App::PropertyLength": _PropertyEditorLength,
    "App::PropertyPercent": _PropertyEditorPercent,
    "App::PropertyString": _PropertyEditorString,
}


def Types():
    """Return the types of properties supported."""
    return [t for t in _EditorFactory]


def Editor(obj, prop):
    """Returns an editor class to be used for the given property."""
    factory = _EditorFactory[obj.getTypeIdOfProperty(prop)]
    if factory:
        return factory(obj, prop)
    return None
